;-------------------------------------------
; system definitions and routines
; (C) Copyright by Marek Karcz 2013.
; All rights reserved.
;
; History of revisions:
; 
; * Some summer day in 2013
;     Initial revision.
; * 9/13/2013
;     Random numbers generator.
;-------------------------------------------

VIC   = $d000     ; VIC base address
SCB   = $0288     ; Screen memory bank register
CCOL  = $0286     ; Character color register
MCOL  = $D800     ; Memory of character colors
MSCR  = $0400     ; Memory of screen - start
MESCR = $0400+1000; Memory of screen - end
SCMEM = VIC + 24  ; Screen (Hi) and character set (Lo) memory register.
BCOL  = VIC + 32  ; Border color register
BGCOL = VIC + 33  ; Background color
; cursor codes
CRSDOWN   = 17
CRSUP     = 145
CRSLEFT   = 157
CRSRIGHT  = 29
; screen control codes
CLRHOME   = 19
CLRSCR    = 147
NL        = 13
DELETE    = 20
; kernal
GETIN   = $ffe4     ; kernal GETIN routine address
CHROUT  = $ffd2     ; kernal CHROUT routine address
MSDELAY = $eeb3     ; kernal routine - 1 millisecond delay
LOAD    = $ffd5     ; Load to RAM
                    ; Kernal LOAD is used by BASIC LOAD to load from tape or
                    ; disk into RAM. The result is not relinked as it is with
                    ; BASIC LOAD. Thus, this routine loads RAM from the device 
                    ; without any changes. Keyboard, RS-232, and screen
                    ; return ILLEGAL DEVICE.
                    ; Since commands like LOAD "filename"',1,1 use a device 
                    ; number and name, SETLFS and SETNAM or the equivalent 
                    ; POKEs have to be used before LOAD.
                    ; Before entering LOAD, A holds 0 for LOAD, 1 (or some 
                    ; nonzero value) for VERIFY. LOAD and BASIC'S VERIFY use 
                    ; almost identical routines, except that VERIFY compares 
                    ; bytes rather than storing them in memory. Provided
                    ; that the secondary address is 0, X (low) and Y (high) point
                    ; to the address at which LOAD will start. If it is nonzero,
                    ; the program will be loaded at the start address stored with
                    ; the file to be loaded.
                    ; LOAD uses a vector after X and Y are stored. The routine 
                    ; branches to $F4B8 (disk LOAD) or $F533 (tape LOAD).
                    ; On exit, C set denotes an error. Register A holds the error 
                    ; number (4, 5, 8, or 9); A, X, Y, and SR are all altered by 
                    ; LOAD. X (low) and Y (high) point to the end address plus 
                    ; one following LOAD.
SAVE    = $ffd8     ; kernal SAVE routine, dumps memory to tape or disk
                    ; requires SETLFS and SETNAM to be called 1st.
                    ; SAVE has to specify start and end address (one byte
                    ; past the end of the block to be saved). Acc holds
                    ; a pointer to a ZPG vector that contains start address,
                    ; X,Y contain the end address (lo,hi respectively).
                    ; SAVE uses a vector after the addresses are stored in
                    ; ($C1) and ($AE). After this, the routine branches to $F5FA
                    ; (disk) or $F659 (tape).
                    ; On exit, C set denotes an error. In this case, A holds
                    ; 5, 8, or 9. However, with disks there may be a disk error
                    ; which has to be read from the disk drive error channel. 
                    ; Alters A, X, Y, and SR.
SETLFS  = $ffba     ; Set logical file number, device address and secondary address.
                    ; This is preliminary routine to opening the file.
                    ; Acc - file number, X - device number, Y - secondary address.
SETNAM  = $ffbd     ; Another preliminary routine to opening the file.
                    ; Sets the file name. Acc - length of the file name.
                    ; X, Y - points to the file name buffer (lo, hi).
                    ; If Acc = 0 (acceptable for tape), X,Y are irrelevant.
OPEN    = $ffc0     ; Opens a file. Assuming that the filename and other parameters 
                    ; have been set by using SETNAM and SETLFS. Entry values 
                    ; of A, X, and Y are thus irrelevant. On exit, carry set
                    ; indicates an error; the error number 1, 2, 4, 5, or 8
                    ; returns in A. Alters A, X, Y, and SR.                    
CLOSE   = $ffc3     ; Close a File
                    ; A holds the file number on entry to this routine, which
                    ; closes that file only, deleting its parameters from the file
                    ; tables and decrementing the number-of-files-open location.
                    ; On exit, C is clear. No errors are reported. Alters A, X, Y,
                    ; and SR.
RDTIM   = $ffde     ; Read Jiffy Clock
                    ; The converse of the previous routine, RDTIM loads Y
                    ; (highest), X (high), and A (low) from the TI clock's bytes.
                    ; The result usually needs some conversion to be useful.
PLOT    = $fff0     ; Cursor Position
                    ; If C flag set: Reads $D6 into X and $D3 into Y, cursor positions 
                    ; down (0-24) and across (0-39), respectively. 
                    ; If Cflag clear: Sets X (down), Y (across). 
                    ; PLOT adjusts the screen links. Alters A, X, and SR. No errors.                    

; ZPG

ZPGA1 = $fb       ; buffer for ZPG indirect addressing/pointer
ZPGA2 = $fd       ; buffer for ZPG indirect addressing/pointer

; routines

!zone SystemZone
; clear screen with kernal routine
.clrscr
    lda #CLRSCR
    jsr CHROUT
    rts
    
; set upper case
.upcaseon
    lda #142
    jsr CHROUT
    rts
    
; disable shift+cbm
.disshcbm
    lda #8
    jsr CHROUT
    rts
    
; enable shift+cbm
.enshftcbm
    lda #9
    jsr CHROUT
    rts

; increment ZPGA1
.inczpga1
    clc
    lda ZPGA1
    adc #1
    sta ZPGA1
    lda ZPGA1+1
    adc #0
    sta ZPGA1+1
    rts
    
; increment ZPGA2
.inczpga2
    clc
    lda ZPGA2
    adc #1
    sta ZPGA2
    lda ZPGA2+1
    adc #0
    sta ZPGA2+1
    rts    
    
; decrement 16-bit counter under ZPGA1 address
.deczpga1
    sec
    lda ZPGA1
    sbc #1
    sta ZPGA1
    lda ZPGA1+1
    sbc #0
    sta ZPGA1+1
    rts
    
; print text pointed by ZPGA1
; any length up to 64 kB
.prnstrzpga1
    ldx #0
    lda (ZPGA1,x)
    beq .prnstr1x
    jsr CHROUT
    jsr .inczpga1
    clc
    bcc .prnstrzpga1
.prnstr1x    
    rts

; get the key pressed or 0 if none in Acc
.getkey
    jsr GETIN
    rts

; fill memory of character colors 
; with code in Acc
.fillmcol

    ldx #0
.l2    
    sta MCOL,x
    sta MCOL+256,x
    sta MCOL+2*256,x
    sta MCOL+3*256,x
    inx
    bne .l2
    rts

; fill screen memory with codes from Acc
.fillscr

    ldx #0
.l3    
    sta MSCR,x
    sta MSCR+256,x
    sta MSCR+2*256,x
    sta MSCR+3*256,x
    inx
    bne .l3
    rts
    
.gpct  !byte 0, 0    
; time delay, A - quarters of a second
; e.g: A = 4, delay = 1 sec
.delay
    sta .gpct+1
.delloop    
    lda .gpct+1
    beq .delend
    dec .gpct+1
    jsr .del4
    clc
    bcc .delloop
.delend    
    rts
    
; delay 0.25 sec
.del4
    lda #250
    sta .gpct
.del4loop
    lda .gpct
    beq .del4end
    dec .gpct
    jsr MSDELAY
    clc
    bcc .del4loop
.del4end
    rts    
    
; copy memory area pointed by ZPGA1 to location pointed by ZPGA2
; the number of bytes to be copied: X - lo, Y - hi
; note: uses gpct counter
.memcpy
    stx .gpct
    sty .gpct+1
    ldx #0
.memcploop
    lda .gpct
    bne .memcpl1
    lda .gpct+1
    bne .memcpl1
    rts
.memcpl1    
    lda (ZPGA1,x)
    sta (ZPGA2,x)
    jsr .decgpct
    jsr .inczpga1
    jsr .inczpga2
    clc
    bcc .memcploop
    
; decrement gpct    
.decgpct
    sec
    lda .gpct
    sbc #1
    sta .gpct
    lda .gpct+1
    sbc #0
    sta .gpct+1
    rts 

.gpbyt !byte 0
; fill memory with a single value
; ZPGA1 - pointer to memory area
; X, Y - number of bytes (lo,hi)
; A - fill value
; note: uses gpct and gpbyt buffers
.fillmem
    stx .gpct
    sty .gpct+1
    sta .gpbyt
    ldx #0
.fillmloop
    lda .gpct
    bne .fillml1
    lda .gpct+1
    bne .fillml1
    rts
.fillml1
    lda .gpbyt
    sta (ZPGA1,x)
    jsr .decgpct
    jsr .inczpga1
    clc
    bcc .fillmloop 
    
; Linear congruential pseudo-random number generator
; by Bruce Clark, 7 Jun 2004
;
; Calculate SEED = 1664525 * SEED + 1
; 
; Enter with:
;
;   SEED0 = byte 0 of seed
;   SEED1 = byte 1 of seed
;   SEED2 = byte 2 of seed
;   SEED3 = byte 3 of seed
;
; Returns:
;
;   SEED0 = byte 0 of seed
;   SEED1 = byte 1 of seed
;   SEED2 = byte 2 of seed
;   SEED3 = byte 3 of seed
;
; TMP, TMP+1, TMP+2 and TMP+3 are overwritten
;
; Assuming that (a) SEED0 to SEED3 and TMP+0 to TMP+3 are all located on page
; zero, and (b) none of the branches cross a page boundary:
;
;   Space: 106 bytes
;   Speed: JSR RAND takes 517 cycles
;
.SEED0   !byte 0
.SEED1   !byte 0
.SEED2   !byte 0
.SEED3   !byte 0
.TMP     !byte 0,0,0,0
.RAND4   !byte  $01,$01,$00,$FE,$FF,$01    
.RAND    CLC          ; copy SEED into TMP
         LDA .SEED0   ; and compute SEED = SEED * $10000 + SEED + 1
         STA .TMP
         ADC #1
         STA .SEED0
         LDA .SEED1
         STA .TMP+1
         ADC #0
         STA .SEED1
         LDA .SEED2
         STA .TMP+2
         ADC .TMP
         STA .SEED2
         LDA .SEED3
         STA .TMP+3
         ADC .TMP+1
         STA .SEED3
;
; Bit 7 of $00, $19, $66, and $0D is 0, so only 6 shifts are necessary
;
         LDY #5
.RAND1   ASL .TMP     ; shift TMP (old seed) left
         ROL .TMP+1
         ROL .TMP+2
         ROL .TMP+3
;
; Get X from the RAND4 table.  When:
;
; X = $00, SEED = SEED + $10000 * TMP
; X = $01, SEED = SEED + $100 * TMP
; X = $FE, SEED = SEED + $10000 * TMP + TMP
; X = $FF, SEED = SEED + $100 * TMP + TMP
;
         LDX .RAND4,Y
         BPL .RAND2   ; branch if X = $00 or X = $01
         CLC          ; SEED = SEED + TMP
         LDA .SEED0
         ADC .TMP
         STA .SEED0
         LDA .SEED1
         ADC .TMP+1
         STA .SEED1
         LDA .SEED2
         ADC .TMP+2
         STA .SEED2
         LDA .SEED3
         ADC .TMP+3
         STA .SEED3
         INX         ; $FE -> $00, $FF -> $01
         INX
.RAND2   CLC
         BEQ .RAND3   ; if X = $00, SEED = SEED + TMP * $10000
         LDA .SEED1   ; SEED = SEED + TMP * $100
         ADC .TMP
         STA .SEED1
.RAND3   LDA .SEED2
         ADC .TMP,X
         STA .SEED2
         LDA .SEED3
         ADC .TMP+1,X
         STA .SEED3
         DEY
         BPL .RAND1
         RTS
